@@ -13,6 +13,7 @@ module Network.Wreq.Internal.Lens
     , requestHeaders
     , requestBody
     , requestVersion
+    , requestManagerOverride
     , onRequestBodyException
     , proxy
     , hostAddress
@@ -25,6 +26,7 @@ module Network.Wreq.Internal.Lens
     , seshCookies
     , seshManager
     , seshRun
+    , seshRunHistory
     -- * Useful functions
     , assoc
     , assoc2
